@@ -1,15 +1,17 @@
 """semopy: Structural Equation Modeling Optimization in Python"""
-from .regularization import create_regularization
+from .model_generalized_effects import ModelGeneralizedEffects
 from .model_effects import ModelEffects
 from .model_means import ModelMeans
+from .regularization import create_regularization
 from .stats import calc_stats, gather_statistics
 from .means import estimate_means
 from .plot import semplot
 from .model import Model
+from . import effects 
 from . import multigroup
 from . import examples
 from . import efa
 
 name = "semopy"
-__version__ = "2.0.20"
+__version__ = "2.1.0"
 __author__ = "Georgy Meshcheryakov"
\ No newline at end of file
